%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% FI:PA171 - Digital Filters
%
% topic: (orthogonal) Image Transforms
% author: David Svoboda
%         Masaryk University (c) 2019
% note: designed for Matlab and DipImage
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Try the following functions:
%
% fft, dct, hadamard, eig
%
% Given a simple highly correlated signal (see below) let us 
% decorrelate it using preselected image transforms. Finally, measure the
% energy compaction of each method. 
%
% The energy of the transformed signal is usually compacted in the first
% element. The PCA is an exception. The element with the greatest energy
% (variance) corresponds to the greatest eigenvalue.

% Create the vector 'f'
f = [1 2 3 5 22 22 24 25 10 11 13 13 -2 -1 -1 1] 

% Choose the constants 'n' and 'k' that split the vector into pieces.
% Note that the following must be valid n*k = length(f).
k = 2
n = length(f)/k

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PCA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear b b_avg V C_V A C_W W

% make 'k' blocks with 'n' pixels each
for i=1:k
    b(i,:) = f((i-1)*n+1 : (i-1)*n+n);
end

% show blocks
b

b_avg = mean(b,1)

for i=1:k
    V(:,i) = b(i,:) - b_avg;
end

% show matrix "V"
V

% make the variance matrix "C_V"
C_V = V*V'

% Find eigenvalues and eigenvectors.
[A,C_W]=eig(C_V)

% We must transpose A as in Matlab the eigenvectors form the columns of A
A = A'

% apply transform & show the results
W = A*V

% find the greatest eigenvalue 
[max_val, max_pos] = max(max(C_W));

% evaluate the energy
for i=1:k
    i
    W(:,i)
    pca_energy = W(:,i)'*W(:,i);
    max_pos
    pca_compactness = (W(max_pos,i)^2 / pca_energy) * 100
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for i=1:k
    i
    fft_solution = (1/sqrt(n)) * fft(f((i-1)*n+1 : (i-1)*n+n));
    transpose(fft_solution)
    fft_energy = fft_solution*fft_solution';
    [max_val, max_pos] = max(abs(fft_solution));
    max_pos
    fft_compactness = (fft_solution(max_pos)^2 / fft_energy) * 100
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for i=1:k
    i
    dct_solution = dct(f((i-1)*n+1 : (i-1)*n+n));
    transpose(dct_solution)
    dct_energy = dct_solution*dct_solution';
    [max_val, max_pos] = max(abs(dct_solution));
    max_pos
    dct_compactness = (dct_solution(max_pos)^2 / dct_energy) * 100
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Hadamard
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

A = (1/sqrt(n))*hadamard(n);
for i=1:k
    i
    hadamard_solution = (A*f((i-1)*n+1 : (i-1)*n+n)')
    hadamard_energy = hadamard_solution'*hadamard_solution;
    [max_val, max_pos] = max(abs(hadamard_solution));
    max_pos
    hadamard_compactness = (hadamard_solution(max_pos)^2 / hadamard_energy) * 100
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Task to solve
% 
% >> Implement your own Fast DCT <<
% 
% Expected syntax:
% function output = my_fdct(input)
%
% Some hints:
% You can assume that length of input is power of 2.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Task to solve (PCA-based information loss):
% 
% (1) Split the into regular rectangular blocks  (DONE)
% (2) The block components should be analyzed using PCA in terms of 
%     their mutual correlation.
% (3) The blocks should be transformed into new space
% (4) Less important components shoule removed/cleared
% (5) The updated component shoule be transformed back to the original
%     vector space.
% (6) The blocks are used to form the new image. The new image and the
% original image should be subtracted to see the differences. (DONE)
%
% Complete the code in file 'reduction.m' (items (2)-(5)).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

